from dateutil import relativedelta, parser, rrule
from dateutil.rrule import WEEKLY
from pathlib import Path
import json

from django.http import HttpResponse, HttpRequest, HttpResponseRedirect, Http404
from django.shortcuts import redirect, render
import requests

from .models import *
from .forms import WhoopAuthForm


def pull_api(user, uri: str) -> dict:
    auth_code = user.whoopuser.access_token
    headers = {'authorization': f'bearer {auth_code}'}
    pull = requests.get(uri, headers=headers)
    # pull.raise_for_status()
    if pull.status_code == 404:
        return dict()
    elif pull.status_code != 200:
        print(pull.status_code)
        return dict()
    elif len(pull.text) == 0:
        print(pull.status_code)
        return dict()
    else:
        return pull.json() or dict()


def pull_sleep_detail(user, sleep_id):
    whoop_user_id = user.whoopuser.whoop_user_id
    uri = f'https://api-7.whoop.com/users/{whoop_user_id}/sleeps/{sleep_id}'
    return pull_api(user, uri)

def get_weekly_ranges(start_date, end_date):
    end_time = 'T23:59:59.999Z'
    start_time = 'T00:00:00.000Z'
    intervals = rrule.rrule(
        freq = WEEKLY,
        interval = 1,
        until = end_date,
        dtstart = start_date
    )
    date_ranges = [
        [
            d.strftime('%Y-%m-%d') + start_time,
            (d + relativedelta.relativedelta(weeks = 1)).strftime('%Y-%m-%d') + end_time
        ]
        for d in intervals
    ]
    return date_ranges


def pull_daily(dates, user, save_json=False):
    cycle_uri = f'https://api-7.whoop.com/users/{user.whoopuser.whoop_user_id}/cycles?end={dates[1]}&start={dates[0]}'
    data = pull_api(user, cycle_uri)

    if save_json:
        Path(f'daily-{dates[0]}--{dates[1]}.json').write_text(json.dumps(data))

    # stuff it all into the database
    for day in data:
        Daily.fill_from_cycle_response(user, day)


def pull_hr(dates, user, save_json=False):
    uri = f'https://api-7.whoop.com/users/{user.whoopuser.whoop_user_id}/metrics/heart_rate?end={dates[1]}&order=t&start={dates[0]}&step=6'
    data = pull_api(user, uri)
    hr_vals = data['values']

    if save_json:
        Path(f'hr-{dates[0]}--{dates[1]}.json').write_text(json.dumps(data))

    HR.fill_from_response(user=user, d=hr_vals)


def loop_historical(user, pull_fun):
    weekly_date_ranges = get_weekly_ranges(
        start_date = user.whoopuser.whoop_createdAt,
        end_date = datetime.datetime.now().replace(tzinfo=pytz.timezone('US/Eastern'))
    )
    for dates in weekly_date_ranges:
        print(f"{dates=}")
        pull_fun(dates, user)


def pull_journal(user, cycle_id):
    uri = f'https://api-7.whoop.com/activities-service/v1/journals/entries/user/cycle/{cycle_id}'
    d = pull_api(user, uri)
    for entry in d.get('tracked_behaviors') or {}:
        JournalEntry.create_from_response(cycle_id, entry)


def pull_all_journals(user):
    cycle_ids = {x.id for x in user.whoopuser.daily_set.all()}

    for cycle_id in cycle_ids:
        pull_journal(user, cycle_id)


def login(request: HttpRequest):
    if request.method == 'POST':
        try:
            whoopUser = WhoopUser.createWithPassword(
                loggedInUser=request.user,
                username=request.POST['username'],
                password=request.POST['password']
            )
            whoopUser.save()
            return redirect('whoopsuccess')
        except Exception as e:
            form = WhoopAuthForm()
            return render(request, 'whoop/login.html', {
                'form': form,
                'error': f'Authentication failed: {str(e)}'
            })
    form = WhoopAuthForm()
    return render(request, 'whoop/login.html', {'form': form})


def reauth(request: HttpRequest):
    if request.method == 'POST':
        try:
            request.user.whoopuser.refreshWithPassword(
                username=request.POST['username'],
                password=request.POST['password']
            )
            return redirect('whoopsuccess')
        except Exception as e:
            form = WhoopAuthForm()
            return render(request, 'whoop/reauth.html', {
                'form': form,
                'error': f'Re-authentication failed: {str(e)}'
            })
    form = WhoopAuthForm()
    return render(request, 'whoop/reauth.html', {'form': form})


def success(request: HttpRequest) -> HttpResponseRedirect:
    # fire off the refresh of all data
    return render(request, "whoop/success.html")


def dashboard(request: HttpRequest):
    """Display WHOOP dashboard with user's data."""
    if not hasattr(request.user, 'whoopuser'):
        return redirect('whooplogin')

    whoopuser = request.user.whoopuser

    # Get recent data (last 30 days)
    thirty_days_ago = datetime.date.today() - datetime.timedelta(days=30)
    recent_days = Daily.objects.filter(
        user=whoopuser,
        day__gte=thirty_days_ago
    ).order_by('-day')[:30]

    # Get summary statistics
    stats = {
        'total_days': Daily.objects.filter(user=whoopuser).count(),
        'avg_recovery': Recovery.objects.filter(day__user=whoopuser).aggregate(
            avg=models.Avg('score')
        )['avg'],
        'avg_sleep_score': Sleep.objects.filter(day__user=whoopuser).aggregate(
            avg=models.Avg('score')
        )['avg'],
        'avg_strain': Strain.objects.filter(day__user=whoopuser).aggregate(
            avg=models.Avg('score')
        )['avg'],
        'total_workouts': Workout.objects.filter(strain__day__user=whoopuser).count(),
    }

    context = {
        'whoopuser': whoopuser,
        'recent_days': recent_days,
        'stats': stats,
    }

    return render(request, 'whoop/dashboard.html', context)


def sync_recent(request: HttpRequest):
    """Sync recent WHOOP data (last 7 days)."""
    if not hasattr(request.user, 'whoopuser'):
        return redirect('whooplogin')

    # Refresh token if needed
    request.user.whoopuser.refreshIfNeeded()

    # Pull last 7 days
    days_back = 7
    days_back_dt = datetime.date.today() - datetime.timedelta(days=days_back)
    dates = [
        days_back_dt.strftime('%Y-%m-%dT00:00:00.000Z'),
        datetime.date.today().strftime('%Y-%m-%dT23:59:59.999Z')
    ]

    try:
        pull_daily(dates, request.user)
        pull_hr(dates, request.user)

        # Pull journal entries for synced days
        cycle_ids = {x.id for x in request.user.whoopuser.daily_set.filter(day__gte=days_back_dt)}
        for cycle_id in cycle_ids:
            pull_journal(request.user, cycle_id)

        return render(request, 'whoop/sync_success.html', {
            'message': f'Successfully synced last {days_back} days of data.',
            'days': days_back
        })
    except Exception as e:
        return render(request, 'whoop/sync_error.html', {
            'error': str(e)
        })


def sync_historical(request: HttpRequest):
    """Sync all historical WHOOP data."""
    if not hasattr(request.user, 'whoopuser'):
        return redirect('whooplogin')

    # Refresh token if needed
    request.user.whoopuser.refreshIfNeeded()

    try:
        loop_historical(request.user, pull_daily)
        loop_historical(request.user, pull_hr)
        pull_all_journals(request.user)

        return render(request, 'whoop/sync_success.html', {
            'message': 'Successfully synced all historical data.',
            'days': 'all'
        })
    except Exception as e:
        return render(request, 'whoop/sync_error.html', {
            'error': str(e)
        })


def data_recovery(request: HttpRequest):
    """View recovery data."""
    if not hasattr(request.user, 'whoopuser'):
        return redirect('whooplogin')

    whoopuser = request.user.whoopuser
    recoveries = Recovery.objects.filter(
        day__user=whoopuser
    ).select_related('day').order_by('-day__day')[:30]

    return render(request, 'whoop/data_recovery.html', {
        'recoveries': recoveries
    })


def data_sleep(request: HttpRequest):
    """View sleep data."""
    if not hasattr(request.user, 'whoopuser'):
        return redirect('whooplogin')

    whoopuser = request.user.whoopuser
    sleeps = Sleep.objects.filter(
        day__user=whoopuser
    ).select_related('day').order_by('-day__day')[:30]

    return render(request, 'whoop/data_sleep.html', {
        'sleeps': sleeps
    })


def data_strain(request: HttpRequest):
    """View strain data."""
    if not hasattr(request.user, 'whoopuser'):
        return redirect('whooplogin')

    whoopuser = request.user.whoopuser
    strains = Strain.objects.filter(
        day__user=whoopuser
    ).select_related('day').order_by('-day__day')[:30]

    return render(request, 'whoop/data_strain.html', {
        'strains': strains
    })


def data_workouts(request: HttpRequest):
    """View workout data."""
    if not hasattr(request.user, 'whoopuser'):
        return redirect('whooplogin')

    whoopuser = request.user.whoopuser
    workouts = Workout.objects.filter(
        strain__day__user=whoopuser
    ).select_related('strain__day').order_by('-during_lower')[:50]

    return render(request, 'whoop/data_workouts.html', {
        'workouts': workouts
    })



"""
Tests for django_whoop views.

The views extend a 'base.html' template that lives in the consuming project,
not inside django_whoop itself.  We inject a stub via Django's locmem loader
using the pytest-django ``settings`` fixture.

WHOOP API calls (requests.post / requests.get) are mocked throughout.
"""
import pytest
from unittest.mock import MagicMock, patch

from django.contrib.auth.models import User
from django.test import Client

from django_whoop.models import WhoopUser


# ---------------------------------------------------------------------------
# Stub base template injected via the ``settings`` fixture
# ---------------------------------------------------------------------------

STUB_BASE = "{% block body %}{% endblock %}"

PATCHED_TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "OPTIONS": {
            "context_processors": ["django.template.context_processors.request"],
            "loaders": [
                (
                    "django.template.loaders.locmem.Loader",
                    {"base.html": STUB_BASE},
                ),
                "django.template.loaders.app_directories.Loader",
            ],
        },
    }
]


@pytest.fixture
def tmpl_settings(settings):
    """Override TEMPLATES to include a stub base.html."""
    settings.TEMPLATES = PATCHED_TEMPLATES


@pytest.fixture
def logged_in_user(db):
    return User.objects.create_user(username="viewer", password="pass")


@pytest.fixture
def logged_in_client(logged_in_user):
    c = Client()
    c.force_login(logged_in_user)
    return c


# ---------------------------------------------------------------------------
# URL routing
# ---------------------------------------------------------------------------

@pytest.mark.django_db
class TestURLRouting:
    def test_login_url_resolves(self):
        from django.urls import reverse
        assert reverse("whooplogin") == "/whoop/login"

    def test_reauth_url_resolves(self):
        from django.urls import reverse
        assert reverse("whoopreauth") == "/whoop/reauth"

    def test_logout_url_resolves(self):
        from django.urls import reverse
        assert reverse("whooplogout") == "/whoop/logout"

    def test_success_url_resolves(self):
        from django.urls import reverse
        assert reverse("whoopsuccess") == "/whoop/success"


# ---------------------------------------------------------------------------
# login view
# ---------------------------------------------------------------------------

@pytest.mark.django_db
class TestLoginView:
    def test_get_returns_200(self, tmpl_settings, logged_in_client):
        response = logged_in_client.get("/whoop/login")
        assert response.status_code == 200

    def test_get_contains_form(self, tmpl_settings, logged_in_client):
        response = logged_in_client.get("/whoop/login")
        assert b"username" in response.content.lower()
        assert b"password" in response.content.lower()

    @patch("django_whoop.models.requests.post")
    def test_post_redirects_to_success(self, mock_post, tmpl_settings, logged_in_client):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = {
            "access_token": "tok123",
            "expires_in": 3600,
            "refresh_token": "ref123",
            "user": {"id": 77, "createdAt": "2020-10-18T00:57:18.609Z"},
        }
        mock_post.return_value = mock_resp

        response = logged_in_client.post(
            "/whoop/login",
            {"username": "whoop@example.com", "password": "secret"},
        )
        assert response.status_code == 302
        assert response["Location"] == "/whoop/success"

    @patch("django_whoop.models.requests.post")
    def test_post_saves_token_to_db(self, mock_post, tmpl_settings, logged_in_client, logged_in_user):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = {
            "access_token": "saved_tok",
            "expires_in": 3600,
            "refresh_token": "saved_ref",
            "user": {"id": 88, "createdAt": "2020-01-01T00:00:00.000Z"},
        }
        mock_post.return_value = mock_resp

        logged_in_client.post(
            "/whoop/login",
            {"username": "w@example.com", "password": "pw"},
        )
        wu = WhoopUser.objects.get(user=logged_in_user)
        assert wu.access_token == "saved_tok"
        assert wu.whoop_user_id == 88


# ---------------------------------------------------------------------------
# reauth view
# ---------------------------------------------------------------------------

@pytest.mark.django_db
class TestReauthView:
    def _make_whoop_user(self, django_user):
        return WhoopUser.objects.create(
            user=django_user,
            access_token="old_tok",
            refresh_token="old_ref",
            whoop_user_id=42,
        )

    def test_get_returns_200(self, tmpl_settings, logged_in_client, logged_in_user):
        self._make_whoop_user(logged_in_user)
        response = logged_in_client.get("/whoop/reauth")
        assert response.status_code == 200

    def test_get_contains_form(self, tmpl_settings, logged_in_client, logged_in_user):
        self._make_whoop_user(logged_in_user)
        response = logged_in_client.get("/whoop/reauth")
        assert b"username" in response.content.lower()

    @patch("django_whoop.models.requests.post")
    def test_post_updates_token_and_redirects(self, mock_post, tmpl_settings, logged_in_client, logged_in_user):
        wu = self._make_whoop_user(logged_in_user)
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = {
            "access_token": "new_tok",
            "expires_in": 3600,
            "refresh_token": "new_ref",
        }
        mock_post.return_value = mock_resp

        response = logged_in_client.post(
            "/whoop/reauth",
            {"username": "w@example.com", "password": "newpw"},
        )
        assert response.status_code == 302
        assert response["Location"] == "/whoop/success"

        wu.refresh_from_db()
        assert wu.access_token == "new_tok"


# ---------------------------------------------------------------------------
# success view
# ---------------------------------------------------------------------------

@pytest.mark.django_db
class TestSuccessView:
    def test_get_returns_200(self, tmpl_settings, logged_in_client):
        response = logged_in_client.get("/whoop/success")
        assert response.status_code == 200

    def test_content(self, tmpl_settings, logged_in_client):
        response = logged_in_client.get("/whoop/success")
        assert b"WHOOP successfully connected" in response.content



"""
Management command to sync WHOOP data.

Usage:
    python manage.py sync_whoop --username <username> --recent
    python manage.py sync_whoop --username <username> --historical
    python manage.py sync_whoop --username <username> --days 30
"""

import datetime
from django.core.management.base import BaseCommand, CommandError
from django.contrib.auth.models import User

from django_whoop.models import WhoopUser
from django_whoop.views import (
    pull_daily, pull_hr, pull_journal, pull_all_journals,
    loop_historical, get_weekly_ranges
)


class Command(BaseCommand):
    help = 'Sync WHOOP data for a user'

    def add_arguments(self, parser):
        parser.add_argument(
            '--username',
            type=str,
            required=True,
            help='Django username to sync data for'
        )
        parser.add_argument(
            '--recent',
            action='store_true',
            help='Sync recent data (last 7 days)'
        )
        parser.add_argument(
            '--historical',
            action='store_true',
            help='Sync all historical data'
        )
        parser.add_argument(
            '--days',
            type=int,
            help='Number of days back to sync'
        )

    def handle(self, *args, **options):
        username = options['username']

        try:
            user = User.objects.get(username=username)
        except User.DoesNotExist:
            raise CommandError(f'User "{username}" does not exist')

        if not hasattr(user, 'whoopuser'):
            raise CommandError(f'User "{username}" does not have a WHOOP account linked')

        whoopuser = user.whoopuser

        # Refresh token if needed
        self.stdout.write(self.style.SUCCESS('Checking token validity...'))
        whoopuser.refreshIfNeeded()

        if options['historical']:
            self.stdout.write(self.style.SUCCESS('Syncing all historical data...'))
            try:
                loop_historical(user, pull_daily)
                self.stdout.write(self.style.SUCCESS('✓ Daily data synced'))

                loop_historical(user, pull_hr)
                self.stdout.write(self.style.SUCCESS('✓ Heart rate data synced'))

                pull_all_journals(user)
                self.stdout.write(self.style.SUCCESS('✓ Journal entries synced'))

                self.stdout.write(self.style.SUCCESS('Historical sync completed successfully!'))
            except Exception as e:
                raise CommandError(f'Error syncing historical data: {str(e)}')

        elif options['recent'] or options.get('days'):
            days_back = options.get('days', 7)
            self.stdout.write(self.style.SUCCESS(f'Syncing last {days_back} days...'))

            try:
                days_back_dt = datetime.date.today() - datetime.timedelta(days=days_back)
                dates = [
                    days_back_dt.strftime('%Y-%m-%dT00:00:00.000Z'),
                    datetime.date.today().strftime('%Y-%m-%dT23:59:59.999Z')
                ]

                pull_daily(dates, user)
                self.stdout.write(self.style.SUCCESS('✓ Daily data synced'))

                pull_hr(dates, user)
                self.stdout.write(self.style.SUCCESS('✓ Heart rate data synced'))

                # Pull journal entries for synced days
                cycle_ids = {x.id for x in whoopuser.daily_set.filter(day__gte=days_back_dt)}
                for cycle_id in cycle_ids:
                    pull_journal(user, cycle_id)
                self.stdout.write(self.style.SUCCESS('✓ Journal entries synced'))

                self.stdout.write(self.style.SUCCESS(f'Successfully synced last {days_back} days!'))
            except Exception as e:
                raise CommandError(f'Error syncing data: {str(e)}')
        else:
            raise CommandError('Please specify --recent, --historical, or --days <n>')



# Generated by Django 3.1.7 on 2021-03-19 16:21

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('whoop', '0009_hr'),
    ]

    operations = [
        migrations.CreateModel(
            name='JournalEntry',
            fields=[
                ('tracker_id', models.IntegerField()),
                ('created_at', models.DateTimeField(max_length=200, null=True)),
                ('updated_at', models.DateTimeField(max_length=200, null=True)),
                ('title', models.CharField(max_length=200)),
                ('status', models.CharField(max_length=200, null=True)),
                ('category', models.CharField(max_length=200, null=True)),
                ('description', models.CharField(max_length=200, null=True)),
                ('question_text', models.CharField(max_length=200, null=True)),
                ('time_label', models.CharField(max_length=200, null=True)),
                ('time_context_label', models.CharField(max_length=200, null=True)),
                ('magnitude', models.JSONField(null=True)),
                ('default_tracker', models.BooleanField(null=True)),
                ('id', models.IntegerField(primary_key=True, serialize=False)),
                ('journal_entry_id', models.IntegerField(null=True)),
                ('entry_created_at', models.DateTimeField(null=True)),
                ('entry_updated_at', models.DateTimeField(null=True)),
                ('answered_yes', models.BooleanField(null=True)),
                ('behavior_tracker_id', models.IntegerField(null=True)),
                ('time_input_value', models.DateTimeField(null=True)),
                ('time_input_label', models.CharField(max_length=200, null=True)),
                ('magnitude_input_label', models.CharField(max_length=200, null=True)),
                ('magnitude_input_value', models.IntegerField(null=True)),
                ('day', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='whoop.daily')),
            ],
        ),
    ]



# Generated by Django 3.1.7 on 2021-03-17 19:24

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('whoop', '0002_auto_20210317_1359'),
    ]

    operations = [
        migrations.AddField(
            model_name='whoopuser',
            name='access_token_expires_in',
            field=models.IntegerField(default=0),
        ),
        migrations.AddField(
            model_name='whoopuser',
            name='access_token_updated',
            field=models.DateTimeField(default=datetime.datetime(1, 1, 1, 0, 0)),
        ),
    ]



from django import forms
from django.utils.translation import gettext_lazy as _
from django.contrib.auth.forms import UsernameField


class WhoopAuthForm(forms.Form):
    """
    Base class for authenticating a whoop user by username/password.
    """
    username = UsernameField(widget=forms.TextInput(attrs={'autofocus': True}))
    password = forms.CharField(
        label=_("Password"),
        strip=False,
        widget=forms.PasswordInput(attrs={'autocomplete': 'current-password'}),
    )



# Generated by Django 3.1.7 on 2021-03-18 19:16

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('whoop', '0007_auto_20210318_1400'),
    ]

    operations = [
        migrations.RenameField(
            model_name='workout',
            old_name='zone_6',
            new_name='zone_0',
        ),
    ]



# Generated by Django 3.1.7 on 2021-03-17 18:59

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('whoop', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='whoopuser',
            name='whoop_createdAt',
            field=models.DateTimeField(default=datetime.datetime(1, 1, 1, 0, 0)),
        ),
        migrations.AddField(
            model_name='whoopuser',
            name='whoop_user_id',
            field=models.IntegerField(default=-1),
        ),
    ]



from django.urls import path

from . import views

urlpatterns = [
    # Authentication
    path('login', views.login, name='whooplogin'),
    path('reauth', views.reauth, name='whoopreauth'),
    path('logout', views.login, name='whooplogout'),
    path('success', views.success, name='whoopsuccess'),

    # Dashboard
    path('', views.dashboard, name='whoopdashboard'),
    path('dashboard', views.dashboard, name='whoopdashboard_alt'),

    # Data syncing
    path('sync/recent', views.sync_recent, name='whoopsyncrecent'),
    path('sync/historical', views.sync_historical, name='whoopsynchistorical'),

    # Data views
    path('data/recovery', views.data_recovery, name='whooprecovery'),
    path('data/sleep', views.data_sleep, name='whoopsleep'),
    path('data/strain', views.data_strain, name='whoopstrain'),
    path('data/workouts', views.data_workouts, name='whoopworkouts'),
]



# Generated by Django 3.1.7 on 2021-03-18 19:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('whoop', '0006_auto_20210318_1359'),
    ]

    operations = [
        migrations.AddField(
            model_name='workout',
            name='during_bounds',
            field=models.CharField(max_length=20, null=True),
        ),
        migrations.AddField(
            model_name='workout',
            name='timezoneOffset',
            field=models.CharField(max_length=10, null=True),
        ),
    ]



# Generated by Django 3.1.7 on 2021-03-18 18:33

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('whoop', '0004_auto_20210317_1430'),
    ]

    operations = [
        migrations.CreateModel(
            name='Daily',
            fields=[
                ('id', models.IntegerField(primary_key=True, serialize=False)),
                ('day', models.DateField()),
                ('during_bounds', models.CharField(max_length=10, null=True)),
                ('during_lower', models.DateTimeField(null=True)),
                ('during_upper', models.DateTimeField(null=True)),
                ('lastUpdatedAt', models.DateTimeField(null=True)),
                ('predictedEnd', models.DateTimeField(null=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='whoop.whoopuser')),
            ],
        ),
        migrations.CreateModel(
            name='Sleep',
            fields=[
                ('id', models.IntegerField(primary_key=True, serialize=False)),
                ('needBreakdown_baseline', models.IntegerField(null=True)),
                ('needBreakdown_debt', models.IntegerField(null=True)),
                ('needBreakdown_naps', models.IntegerField(null=True)),
                ('needBreakdown_strain', models.IntegerField(null=True)),
                ('needBreakdown_total', models.IntegerField(null=True)),
                ('qualityDuration', models.IntegerField(null=True)),
                ('score', models.IntegerField(null=True)),
                ('day', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='whoop.daily')),
            ],
        ),
        migrations.CreateModel(
            name='Strain',
            fields=[
                ('id', models.IntegerField(primary_key=True, serialize=False)),
                ('averageHeartRate', models.FloatField(null=True)),
                ('kilojoules', models.FloatField(null=True)),
                ('maxHeartRate', models.FloatField(null=True)),
                ('rawScore', models.FloatField(null=True)),
                ('score', models.FloatField(null=True)),
                ('day', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='whoop.daily')),
            ],
        ),
        migrations.CreateModel(
            name='Workout',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('zone_1', models.IntegerField(null=True)),
                ('zone_2', models.IntegerField(null=True)),
                ('zone_3', models.IntegerField(null=True)),
                ('zone_4', models.IntegerField(null=True)),
                ('zone_5', models.IntegerField(null=True)),
                ('zone_6', models.IntegerField(null=True)),
                ('strain', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='whoop.strain')),
            ],
        ),
        migrations.CreateModel(
            name='SleepDetail',
            fields=[
                ('id', models.IntegerField(primary_key=True, serialize=False)),
                ('cyclesCount', models.IntegerField(null=True)),
                ('disturbanceCount', models.IntegerField(null=True)),
                ('during_bounds', models.CharField(max_length=20, null=True)),
                ('during_lower', models.DateTimeField(null=True)),
                ('during_upper', models.DateTimeField(null=True)),
                ('inBedDuration', models.IntegerField(null=True)),
                ('isNap', models.BooleanField(default=False)),
                ('latencyDuration', models.IntegerField(null=True)),
                ('lightSleepDuration', models.IntegerField(null=True)),
                ('noDataDuration', models.IntegerField(null=True)),
                ('qualityDuration', models.IntegerField(null=True)),
                ('remSleepDuration', models.IntegerField(null=True)),
                ('respiratoryRate', models.FloatField(null=True)),
                ('score', models.IntegerField(null=True)),
                ('sleepConsistency', models.IntegerField(null=True)),
                ('sleepEfficiency', models.FloatField(null=True)),
                ('slowWaveSleepDuration', models.IntegerField(null=True)),
                ('timezoneOffset', models.CharField(max_length=20, null=True)),
                ('wakeDuration', models.IntegerField(null=True)),
                ('sleep', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='whoop.sleep')),
            ],
        ),
        migrations.CreateModel(
            name='Recovery',
            fields=[
                ('heartRateVariabilityRmssd', models.FloatField(null=True)),
                ('id', models.IntegerField(primary_key=True, serialize=False)),
                ('restingHeartRate', models.FloatField(null=True)),
                ('score', models.FloatField(null=True)),
                ('timestamp', models.DateTimeField(null=True)),
                ('day', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='whoop.daily')),
            ],
        ),
    ]



## this helps with weird ipython completion error
# get_ipython().run_line_magic('config', 'IPCompleter.use_jedi = False')

## start django setup

from pathlib import Path
import sys; sys.path.insert(0, str(Path(sys.path[0]) / '../..'))
print(sys.path)
import os
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "main.settings")
# from asgiref.sync import sync_to_async
os.environ["DJANGO_ALLOW_ASYNC_UNSAFE"] = "true"
import django
django.setup()

## completed django setup

import json

from django.contrib.auth.models import User

from django_whoop.models import *
from django_whoop.views import *

user = User.objects.get(username='andyreagan')
user.whoopuser.refreshIfNeeded()


def full_history():
    loop_historical(user, pull_daily)
    loop_historical(user, pull_hr)
    # cycle_id = user.whoopuser.daily_set.all()[100].id
    # pull_journal(user, cycle_id)
    pull_all_journals(user)


def last_n_days(days_back):
    days_back_dt = datetime.date.today() - datetime.timedelta(days=days_back)
    dates = [days_back_dt.strftime('%Y-%m-%dT00:00:00.000Z'), datetime.date.today().strftime('%Y-%m-%dT23:59:59.999Z')]
    print(f"{dates=}")
    pull_daily(dates, user)
    pull_hr(dates, user)

    cycle_ids = {x.id for x in user.whoopuser.daily_set.filter(day__gte=days_back_dt)}
    for cycle_id in cycle_ids:
        pull_journal(user, cycle_id)

last_n_days(3)


"""
Minimal Django settings for running django-whoop tests.
"""

SECRET_KEY = "django-whoop-test-secret-key-not-for-production"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

INSTALLED_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.auth",
    "django.contrib.sessions",
    "django_whoop",
]

SESSION_ENGINE = "django.contrib.sessions.backends.db"

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
]

# django_whoop/apps.py sets label = 'whoop', so migrations live under that label
DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

USE_TZ = True

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        # Include the app's own templates directory so whoop/login.html etc. resolve.
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.request",
            ],
        },
    }
]

ROOT_URLCONF = "tests.urls"



# Generated by Django 3.1.7 on 2021-03-17 19:30

import datetime
import pytz
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('whoop', '0003_auto_20210317_1424'),
    ]

    operations = [
        migrations.AlterField(
            model_name='whoopuser',
            name='access_token_updated',
            field=models.DateTimeField(default=datetime.datetime(1, 1, 1, 0, 0, tzinfo=pytz.UTC)),
        ),
        migrations.AlterField(
            model_name='whoopuser',
            name='whoop_createdAt',
            field=models.DateTimeField(default=datetime.datetime(1, 1, 1, 0, 0, tzinfo=pytz.UTC)),
        ),
    ]



"""
Tests for django_whoop forms.
"""
import pytest
from django_whoop.forms import WhoopAuthForm


class TestWhoopAuthForm:
    def test_valid_data(self):
        form = WhoopAuthForm(data={"username": "user@example.com", "password": "s3cr3t"})
        assert form.is_valid()

    def test_missing_username(self):
        form = WhoopAuthForm(data={"password": "s3cr3t"})
        assert not form.is_valid()
        assert "username" in form.errors

    def test_missing_password(self):
        form = WhoopAuthForm(data={"username": "user@example.com"})
        assert not form.is_valid()
        assert "password" in form.errors

    def test_empty_form_invalid(self):
        form = WhoopAuthForm(data={})
        assert not form.is_valid()

    def test_password_field_is_password_input(self):
        from django.forms import PasswordInput
        form = WhoopAuthForm()
        assert isinstance(form.fields["password"].widget, PasswordInput)

    def test_username_field_has_autofocus(self):
        form = WhoopAuthForm()
        widget_attrs = form.fields["username"].widget.attrs
        assert widget_attrs.get("autofocus") is True



import datetime
import pytz

from django.contrib.auth.models import User
from django.db import models
import requests


def get_date(raw: str, tzinfo=None):
    if raw is None:
        return None
    datetime_formats = ['%Y', '%Y-%m-%d', '%Y-%m-%dT%H:%M:%S.%fZ', '%Y-%m-%dT%H:%M:%S.%f', '%Y-%m-%dT%H:%M:%SZ', '%Y-%m-%dT%H:%M:%S', '%Y-%m-%d %H:%M:%S', "%Y-%m-%dT%H:%M:%S.%f+00:00", "%Y-%m-%dT%H:%M:%S+00:00", '%Y-%m-%d %H:%M:%S.%fZ', '%Y-%m-%d %H:%M:%S.%f', '%Y-%m-%d %H:%M:%SZ', '%Y-%m-%d %H:%M:%S', '%Y-%m-%d %H:%M:%S', "%Y-%m-%d %H:%M:%S.%f+00:00", "%Y-%m-%d %H:%M:%S+00:00"]
    for dt_format in datetime_formats:
        try:
            d = datetime.datetime.strptime(raw, dt_format)
        except ValueError:
            d = None
        if d is not None:
            if tzinfo is not None:
                return d.replace(tzinfo=tzinfo)
            else:
                return d
    return d


def get_utc_time(raw):
    # raw can be int, float, str convertable to float, or null
    # will try to divide by 1000 if it doesn't work out of the box
    if raw is not None:
        try:
            d = datetime.datetime.utcfromtimestamp(float(raw)).replace(tzinfo=pytz.UTC)
        except ValueError:
            d = datetime.datetime.utcfromtimestamp(float(raw)/1000).replace(tzinfo=pytz.UTC)
        return d
    else:
        return None


class WhoopUser(models.Model):
    """Adds a token to a django User."""

    user = models.OneToOneField(User, on_delete=models.CASCADE)
    access_token = models.CharField(max_length=1024)
    access_token_updated = models.DateTimeField(default=datetime.datetime(1, 1, 1, 0, 0).replace(tzinfo=pytz.UTC))
    access_token_expires_in = models.IntegerField(default=0)
    refresh_token = models.CharField(max_length=1024)
    whoop_user_id = models.IntegerField(default=-1)
    whoop_createdAt = models.DateTimeField(default=datetime.datetime(1, 1, 1, 0, 0).replace(tzinfo=pytz.UTC))

    def __str__(self):
        return self.user.username

    @classmethod
    def createWithPassword(cls, loggedInUser, username, password):
        headers = {
            "username": username,
            "password": password,
            "grant_type": "password",
            "issueRefresh": True
        }
        auth = requests.post("https://api-7.whoop.com/oauth/token", json=headers)
        print(auth)
        print(auth.text)
        print(auth.status_code)
        # should check we got a 200...
        resp = auth.json()
        whoopUser = cls(
            user=loggedInUser,
            access_token=resp['access_token'],
            access_token_updated=datetime.datetime.now().replace(tzinfo=pytz.UTC),
            access_token_expires_in=resp['expires_in'],
            refresh_token=resp['refresh_token'],
            whoop_user_id=resp['user']['id'],
            # there are two dates to choose from, the profile is 3 days later
            # in my case (ex: '2020-10-18T00:57:18.609Z')
            # createdAt=resp['user']['profile']['createdAt']
            whoop_createdAt=get_date(resp['user']['createdAt'], tzinfo=pytz.UTC)
        )
        return whoopUser


    def refreshWithPassword(self, username, password):
        headers = {
            "username": username,
            "password": password,
            "grant_type": "password",
            "issueRefresh": True
        }
        auth = requests.post("https://api-7.whoop.com/oauth/token", json=headers)
        print(auth)
        print(auth.text)
        print(auth.status_code)
        # should check we got a 200...
        resp = auth.json()
        self.access_token = resp['access_token']
        self.access_token_updated = datetime.datetime.now().replace(tzinfo=pytz.UTC)
        self.access_token_expires_in = resp['expires_in']
        self.refresh_token=resp['refresh_token']
        self.save()

    def refreshIfNeeded(self, check_api=False):
        """
        Refresh the access token if it has expired.

        Args:
            check_api: If True, make an API call to check token validity.
                      If False, check based on timestamp.
        """
        if check_api:
            # Check token validity by making a test API call
            headers = {'authorization': f'bearer {self.access_token}'}
            test_url = f'https://api-7.whoop.com/users/{self.whoop_user_id}/profile'
            try:
                response = requests.get(test_url, headers=headers)
                if response.status_code == 401:
                    # Token is invalid, refresh it
                    self.refreshToken()
            except requests.RequestException:
                # If API call fails, try refreshing anyway
                self.refreshToken()
        else:
            # Check if token has expired based on timestamp
            # Token expires_in is in seconds
            token_age = (datetime.datetime.now(pytz.UTC) - self.access_token_updated).total_seconds()
            # Refresh if token is older than expires_in (with 5 minute buffer)
            if token_age > (self.access_token_expires_in - 300):
                self.refreshToken()

    def refreshToken(self):
        headers = {
            "refresh_token": self.refresh_token,
            "grant_type": "refresh_token"
        }
        auth = requests.post("https://api-7.whoop.com/oauth/token", json=headers)
        print(auth.status_code)
        resp = auth.json()
        print(resp)
        self.access_token = resp['access_token']
        self.access_token_updated = datetime.datetime.now().replace(tzinfo=pytz.UTC)
        self.access_token_expires_in = resp['expires_in']
        self.save()


class Daily(models.Model):
    id = models.IntegerField(primary_key=True)
    user = models.ForeignKey(WhoopUser, on_delete=models.CASCADE)
    day = models.DateField()

    during_bounds = models.CharField(max_length=10, null=True)
    during_lower = models.DateTimeField(null=True)
    during_upper = models.DateTimeField(null=True)

    lastUpdatedAt = models.DateTimeField(null=True)
    predictedEnd = models.DateTimeField(null=True)

    def __str__(self):
        return self.day.strftime('%Y-%m-%d')

    @classmethod
    def fill_from_cycle_response(cls, user, d):
        assert 'id' in d
        assert 'days' in d and len(d['days']) == 1
        during = d.get('during') or dict()
        daily, created = cls.objects.update_or_create(
            user=user.whoopuser,
            id=d['id'],
            defaults={
                'day': get_date(d['days'][0]).date(),
                'during_bounds': during.get('bounds'),
                'during_lower': get_date(during.get('lower'), tzinfo=pytz.UTC),
                'during_upper': get_date(during.get('upper'), tzinfo=pytz.UTC),
                'lastUpdatedAt': get_date(d.get('lastUpdatedAt'), tzinfo=pytz.UTC),
                'predictedEnd': get_date(d.get('predictedEnd'), tzinfo=pytz.UTC)
            }
        )
        daily.save()
        print(f"{('found', 'created')[created]} the day {daily=}")
        if 'recovery' in d and d['recovery'] is not None:
            recovery, created = Recovery.objects.update_or_create(
                day=daily,
                id=d['recovery'].get('id'),
                defaults={
                    'heartRateVariabilityRmssd': d['recovery'].get('heartRateVariabilityRmssd'),
                    'restingHeartRate': d['recovery'].get('restingHeartRate'),
                    'score': d['recovery'].get('score'),
                    'timestamp': get_date(d['recovery'].get('timestamp'), tzinfo=pytz.UTC)
                }
            )
            recovery.save()
            print(f"{('found', 'created')[created]} {recovery=}")
        if 'sleep' in d and d['sleep'] is not None and d['sleep'].get('id') is not None:
            needBreakdown = d['sleep'].get('needBreakdown') or dict()
            sleep, created = Sleep.objects.update_or_create(
                day=daily,
                id=d['sleep']['id'],
                defaults={
                    'needBreakdown_baseline': needBreakdown.get('baseline'),
                    'needBreakdown_debt': needBreakdown.get('debt'),
                    'needBreakdown_naps': needBreakdown.get('naps'),
                    'needBreakdown_strain': needBreakdown.get('strain'),
                    'needBreakdown_total': needBreakdown.get('total'),
                    'qualityDuration': d['sleep'].get('qualityDuration'),
                    'score': d['sleep'].get('score')
                }
            )
            print(f"{('found', 'created')[created]} {sleep=}")
            sleep.save()
            if d['sleep'].get('naps') is not None:
                for sleepdetail in d['sleep']['naps']:
                    # could do the below with isNap=True
                    pass
            if d['sleep'].get('sleeps') is not None:
                for sleepdetail in d['sleep']['sleeps']:
                    during = sleepdetail.get('during') or dict()
                    if sleepdetail.get('id') is not None:
                        sleepdetail_obj, created = SleepDetail.objects.update_or_create(
                            sleep=sleep,
                            id=sleepdetail['id'], # = models.IntegerField() # 131838178,
                            defaults={
                                'cyclesCount': sleepdetail.get('cyclesCount'), # = models.IntegerField(null=True) # 5,
                                'disturbanceCount': sleepdetail.get('disturbanceCount'), # = models.IntegerField(null=True) # 12,
                                'during_bounds': during.get('bounds'), # = models.CharField(null=True) # "[)",
                                'during_lower': get_date(during.get('lower'), tzinfo=pytz.UTC), # = models.DateTimeField(null=True) # "2020-10-25T03:01:00+00:00",
                                'during_upper': get_date(during.get('upper'), tzinfo=pytz.UTC), # = models.DateTimeField(null=True) # "2020-10-25T11:33:00+00:00",
                                'inBedDuration': sleepdetail.get('inBedDuration'), # = models.IntegerField(null=True) # 30718493,
                                'isNap': False, # = models.BooleanField(default=False) # false,
                                'latencyDuration': sleepdetail.get('latencyDuration'), # = models.IntegerField(null=True) # 0,
                                'lightSleepDuration': sleepdetail.get('lightSleepDuration'), # = models.IntegerField(null=True) # 12011596,
                                'noDataDuration': sleepdetail.get('noDataDuration'), # = models.IntegerField(null=True) # 0,
                                'qualityDuration': sleepdetail.get('qualityDuration'), # = models.IntegerField(null=True) # 25139744,
                                'remSleepDuration': sleepdetail.get('remSleepDuration'), # = models.IntegerField(null=True) # 5948524,
                                'respiratoryRate': sleepdetail.get('respiratoryRate'), # = models.FloatField(null=True) # 14.3262,
                                'score': sleepdetail.get('score'), # = models.IntegerField(null=True) # 73,
                                'sleepConsistency': sleepdetail.get('sleepConsistency'), # = models.IntegerField(null=True) # 89,
                                'sleepEfficiency': sleepdetail.get('sleepEfficiency'), # = models.FloatField(null=True) # 0.818391,
                                'slowWaveSleepDuration': sleepdetail.get('slowWaveSleepDuration'), # = models.IntegerField(null=True) # 7179624,
                                'timezoneOffset': sleepdetail.get('timezoneOffset'), # = models.CharField(null=True) # "-0400",
                                'wakeDuration': sleepdetail.get('wakeDuration') # = models.IntegerField(null=True) # 5615410
                            }
                        )
                        sleepdetail_obj.save()
                        print(f"{('found', 'created')[created]} {sleepdetail_obj=}")
        if 'strain' in d and d['strain'] is not None:
            strain, created = Strain.objects.update_or_create(
                day=daily,
                id=d['id'],
                defaults={
                    'averageHeartRate': d['strain'].get('averageHeartRate'), # models.FloatField(null=True) # 75,
                    'kilojoules': d['strain'].get('kilojoules'), # models.FloatField(null=True) # 14218.7,
                    'maxHeartRate': d['strain'].get('maxHeartRate'), # models.FloatField(null=True) # 172,
                    'rawScore': d['strain'].get('rawScore'), # models.FloatField(null=True) # 0.0143161574422024,
                    'score': d['strain'].get('score') # models.FloatField(null=True) # 16.3900610678999,
                }
            )
            strain.save()
            print(f"{('found', 'created')[created]} {strain=}")
            if d['strain'].get('workouts') is not None:
                for workout in d['strain']['workouts']:
                    during = workout.get('during') or dict()
                    zones = workout.get('zones')
                    if zones is None or len(zones) < 6:
                        zones = [0 for i in range(6)]
                    if workout.get('id') is not None:
                        workout_obj, created = Workout.objects.update_or_create(
                            strain=strain, # = models.ForeignKey(Strain, on_delete=models.CASCADE)
                            id=workout['id'], # = models.IntegerField(null=True), # 132062158,
                            defaults={
                                'averageHeartRate': workout.get('averageHeartRate'), # = models.FloatField(null=True), # 131,
                                'cumulativeWorkoutStrain': workout.get('cumulativeWorkoutStrain'), # = models.FloatField(null=True), # 14.9973,
                                'during_bounds': during.get('bounds'), # = models.CharField(null=True) # "[)",
                                'during_lower': get_date(during.get('lower'), tzinfo=pytz.UTC), # = models.DateTimeField(null=True) # "2020-10-25T03:01:00+00:00",
                                'during_upper': get_date(during.get('upper'), tzinfo=pytz.UTC), # = models.DateTimeField(null=True) # "2020-10-25T11:33:00+00:00",
                                'kilojoules': workout.get('kilojoules'), # = models.FloatField(null=True), # 2863.52,
                                'maxHeartRate': workout.get('maxHeartRate'), # = models.FloatField(null=True), # 157,
                                'rawScore': workout.get('rawScore'), # = models.FloatField(null=True), # 0.00579152500086973,
                                'score': workout.get('score'), # = models.FloatField(null=True), # 12.2451922891289,
                                'sportId': workout.get('sportId'), # = models.IntegerField(null=True), # 0,
                                'timezoneOffset': workout.get('timezoneOffset'), # = models.CharField(max_length=10, null=True), # "-0400",
                                'zone_0': zones[0], # = models.IntegerField(null=True)
                                'zone_1': zones[1], # = models.IntegerField(null=True)
                                'zone_2': zones[2], # = models.IntegerField(null=True)
                                'zone_3': zones[3], # = models.IntegerField(null=True)
                                'zone_4': zones[4], # = models.IntegerField(null=True)
                                'zone_5': zones[5] # = models.IntegerField(null=True)
                            }
                        )
                        workout_obj.save()
                        print(f"{('found', 'created')[created]} {workout_obj=}")
        return daily


class Recovery(models.Model):
    day = models.ForeignKey(Daily, on_delete=models.CASCADE)
    # blackoutUntil = models.FloatField(null=True) # null,
    # calibrating = models.FloatField(null=True) # false,
    heartRateVariabilityRmssd = models.FloatField(null=True) # 0.0590787,
    id = models.IntegerField(primary_key=True) # 60322913,
    # responded = models.FloatField(null=True) # false,
    restingHeartRate = models.FloatField(null=True) # 51,
    score = models.FloatField(null=True) # 54,
    # state = models.FloatField(null=True) # "complete",
    # surveyResponseId = models.FloatField(null=True) # null,
    timestamp = models.DateTimeField(null=True) # "2020-10-25T11:33:00+00:00",


class Sleep(models.Model):
    day = models.ForeignKey(Daily, on_delete=models.CASCADE)
    id = models.IntegerField(primary_key=True)
    needBreakdown_baseline = models.IntegerField(null=True) # 27430627,
    needBreakdown_debt = models.IntegerField(null=True) # 3703134,
    needBreakdown_naps = models.IntegerField(null=True) # 0,
    needBreakdown_strain = models.IntegerField(null=True) # 3278169,
    needBreakdown_total = models.IntegerField(null=True) # 34411932,
    qualityDuration = models.IntegerField(null=True) # 25139744,
    score = models.IntegerField(null=True) # 73,


class SleepDetail(models.Model):
    sleep = models.ForeignKey(Sleep, on_delete=models.CASCADE)
    id = models.IntegerField(primary_key=True) # 131838178,
    cyclesCount = models.IntegerField(null=True) # 5,
    disturbanceCount = models.IntegerField(null=True) # 12,
    during_bounds = models.CharField(max_length=20, null=True) # "[)",
    during_lower = models.DateTimeField(null=True) # "2020-10-25T03:01:00+00:00",
    during_upper = models.DateTimeField(null=True) # "2020-10-25T11:33:00+00:00",
    inBedDuration = models.IntegerField(null=True) # 30718493,
    isNap = models.BooleanField(default=False) # false,
    latencyDuration = models.IntegerField(null=True) # 0,
    lightSleepDuration = models.IntegerField(null=True) # 12011596,
    noDataDuration = models.IntegerField(null=True) # 0,
    qualityDuration = models.IntegerField(null=True) # 25139744,
    remSleepDuration = models.IntegerField(null=True) # 5948524,
    respiratoryRate = models.FloatField(null=True) # 14.3262,
    # responded = models.FloatField(null=True) # false,
    score = models.IntegerField(null=True) # 73,
    sleepConsistency = models.IntegerField(null=True) # 89,
    sleepEfficiency = models.FloatField(null=True) # 0.818391,
    slowWaveSleepDuration = models.IntegerField(null=True) # 7179624,
    # source = models.FloatField(null=True) # "user",
    # state = models.FloatField(null=True) # "complete",
    # surveyResponseId = models.FloatField(null=True) # null,
    timezoneOffset = models.CharField(max_length=20, null=True) # "-0400",
    wakeDuration = models.IntegerField(null=True) # 5615410

class Strain(models.Model):
    day = models.ForeignKey(Daily, on_delete=models.CASCADE)
    id = models.IntegerField(primary_key=True) #
    averageHeartRate = models.FloatField(null=True) # 75,
    kilojoules = models.FloatField(null=True) # 14218.7,
    maxHeartRate = models.FloatField(null=True) # 172,
    rawScore = models.FloatField(null=True) # 0.0143161574422024,
    score = models.FloatField(null=True) # 16.3900610678999,
    # state = models.FloatField(null=True) # "complete",


class Workout(models.Model):
    strain = models.ForeignKey(Strain, on_delete=models.CASCADE)
    # "altitudeChange": null,
    # "altitudeGain": null,
    averageHeartRate = models.FloatField(null=True) # 131,
    cumulativeWorkoutStrain = models.FloatField(null=True) # 14.9973,
    # distance = models.FloatField(null=True) # null,
    during_bounds = models.CharField(max_length=20, null=True) # "[)",
    during_lower = models.DateTimeField(null=True) # "2020-10-25T16:53:55.099+00:00",
    during_upper = models.DateTimeField(null=True) # "2020-10-25T18:08:15.545+00:00",
    # gpsEnabled = models.FloatField(null=True) # false,
    id = models.IntegerField(primary_key=True) # 132062158,
    kilojoules = models.FloatField(null=True) # 2863.52,
    maxHeartRate = models.FloatField(null=True) # 157,
    rawScore = models.FloatField(null=True) # 0.00579152500086973,
    # responded = models.FloatField(null=True) # false,
    score = models.FloatField(null=True) # 12.2451922891289,
    # source = models.FloatField(null=True) # "auto",
    sportId = models.IntegerField(null=True) # 0,
    # state = models.FloatField(null=True) # "complete",
    # surveyResponseId = models.FloatField(null=True) # null,
    timezoneOffset = models.CharField(max_length=10, null=True) # "-0400",
    zone_0 = models.IntegerField(null=True)
    zone_1 = models.IntegerField(null=True)
    zone_2 = models.IntegerField(null=True)
    zone_3 = models.IntegerField(null=True)
    zone_4 = models.IntegerField(null=True)
    zone_5 = models.IntegerField(null=True)


class HR(models.Model):
    user = models.ForeignKey(WhoopUser, on_delete=models.CASCADE)
    timestamp = models.DateTimeField()
    value = models.IntegerField()

    @classmethod
    def fill_from_response(cls, user, d, batch_size=500):
        # for point in d:
        #     hr, created = cls.objects.update_or_create(
        #         user = user.whoopuser,
        #         timestamp = datetime.datetime.utcfromtimestamp(point['time']/1000).replace(tzinfo=pytz.UTC),
        #         defaults = {'value': point['data']}
        #     )
        #     hr.save()

        # bulk create:
        objects = [cls(user = user.whoopuser,timestamp = datetime.datetime.utcfromtimestamp(point['time']/1000).replace(tzinfo=pytz.UTC), value=point['data']) for point in d]
        cls.objects.bulk_create(objects, batch_size)


class JournalEntry(models.Model):
    # user = models.ForeignKey(WhoopUser, on_delete=models.CASCADE)
    # it's pulled using the day's id, as "cycle_id"
    day = models.ForeignKey(Daily, on_delete=models.CASCADE)

    tracker_id = models.IntegerField() # "id": 50,
    created_at = models.DateTimeField(null=True, max_length=200) # "created_at": 1583441766.111,
    updated_at = models.DateTimeField(null=True, max_length=200) # "updated_at": 1585669607.122,
    title = models.CharField(max_length=200) # "title": "Intermittent Fasting",
    status = models.CharField(null=True, max_length=200) # "status": "active",
    category = models.CharField(null=True, max_length=200) # "category": "Nutrition",
    # sticky = models.BooleanField(null=True, max_length=200) # "sticky": false,
    description = models.CharField(null=True, max_length=200) # "description": null,
    question_text = models.CharField(null=True, max_length=200) # "question_text": "Follow an intermittent fasting diet?",
    time_label = models.CharField(null=True, max_length=200) # "time_label": "When was your last meal?",
    time_context_label = models.CharField(null=True, max_length=200) # "time_context_label": "stopped eating at",
    magnitude = models.JSONField(null=True) # "magnitude": ...
    default_tracker = models.BooleanField(null=True) # "default_tracker": false

    id = models.IntegerField(primary_key=True) # "id": 434680106,
    journal_entry_id = models.IntegerField(null=True) # "journal_entry_id": 52336709,
    entry_created_at = models.DateTimeField(null=True) # "created_at": 1611923762.42,
    entry_updated_at = models.DateTimeField(null=True) # "updated_at": 1611923762.42,
    answered_yes = models.BooleanField(null=True) # "answered_yes": false,
    behavior_tracker_id = models.IntegerField(null=True) # "behavior_tracker_id": 50,
    time_input_value = models.DateTimeField(null=True) # "time_input_value": null,
    time_input_label = models.CharField(max_length=200, null=True) # "time_input_label": null,
    magnitude_input_label = models.CharField(max_length=200, null=True) # "magnitude_input_label": null,
    magnitude_input_value = models.IntegerField(null=True) # "magnitude_input_value": null

    def __str__(self):
        return f"{str(self.day.day)} {self.title}. answered_yes: {self.answered_yes}"

    @classmethod
    def create_from_response(cls, cycle_id, d):
        behavior_tracker = d.get('behavior_tracker') or dict()
        tracker_input = d.get('tracker_input') or dict()
        journalentry, created = cls.objects.update_or_create(
            day=Daily.objects.get(id=cycle_id),
            id=tracker_input.get('id'),
            defaults={
                'tracker_id': behavior_tracker.get('id'),
                'created_at': get_utc_time(behavior_tracker.get('created_at')),
                'updated_at': get_utc_time(behavior_tracker.get('updated_at')),
                'title': behavior_tracker.get('title'),
                'status': behavior_tracker.get('status'),
                'category': behavior_tracker.get('category'),
                # 'sticky': behavior_tracker.get('sticky'),
                'description': behavior_tracker.get('description'),
                'question_text': behavior_tracker.get('question_text'),
                'time_label': behavior_tracker.get('time_label'),
                'time_context_label': behavior_tracker.get('time_context_label'),
                'magnitude': behavior_tracker.get('magnitude'),
                'default_tracker': behavior_tracker.get('default_tracker'),
                'journal_entry_id': tracker_input.get('journal_entry_id'),
                'entry_created_at': get_utc_time(tracker_input.get('created_at')),
                'entry_updated_at': get_utc_time(tracker_input.get('updated_at')),
                'answered_yes': tracker_input.get('answered_yes'),
                'behavior_tracker_id': tracker_input.get('behavior_tracker_id'),
                'time_input_value': get_utc_time(tracker_input.get('time_input_value')),
                'time_input_label': tracker_input.get('time_input_label'),
                'magnitude_input_label': tracker_input.get('magnitude_input_label'),
                'magnitude_input_value': tracker_input.get('magnitude_input_value')
            }
        )
        print(f"{('found', 'created')[created]} {journalentry=}")
        journalentry.save()
        return journalentry



# Generated by Django 3.1.7 on 2021-03-18 18:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('whoop', '0005_daily_recovery_sleep_sleepdetail_strain_workout'),
    ]

    operations = [
        migrations.AddField(
            model_name='workout',
            name='averageHeartRate',
            field=models.FloatField(null=True),
        ),
        migrations.AddField(
            model_name='workout',
            name='cumulativeWorkoutStrain',
            field=models.FloatField(null=True),
        ),
        migrations.AddField(
            model_name='workout',
            name='during_lower',
            field=models.DateTimeField(null=True),
        ),
        migrations.AddField(
            model_name='workout',
            name='during_upper',
            field=models.DateTimeField(null=True),
        ),
        migrations.AddField(
            model_name='workout',
            name='kilojoules',
            field=models.FloatField(null=True),
        ),
        migrations.AddField(
            model_name='workout',
            name='maxHeartRate',
            field=models.FloatField(null=True),
        ),
        migrations.AddField(
            model_name='workout',
            name='rawScore',
            field=models.FloatField(null=True),
        ),
        migrations.AddField(
            model_name='workout',
            name='score',
            field=models.FloatField(null=True),
        ),
        migrations.AddField(
            model_name='workout',
            name='sportId',
            field=models.IntegerField(null=True),
        ),
        migrations.AlterField(
            model_name='workout',
            name='id',
            field=models.IntegerField(primary_key=True, serialize=False),
        ),
    ]



"""
Minimal URL config for tests.

django_whoop's templates extend 'base.html' and reference {% url 'home' %},
so we provide a stub base template via a simple view and wire up the whoop
URLs under the same prefix used in a typical integration.
"""
from django.urls import path, include
from django.http import HttpResponse


def stub_home(request):
    return HttpResponse("home")


urlpatterns = [
    path("whoop/", include("django_whoop.urls")),
    path("", stub_home, name="home"),
]



# Generated by Django 3.1.7 on 2021-03-18 19:46

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('whoop', '0008_auto_20210318_1416'),
    ]

    operations = [
        migrations.CreateModel(
            name='HR',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('timestamp', models.DateTimeField()),
                ('value', models.IntegerField()),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='whoop.whoopuser')),
            ],
        ),
    ]



from django.contrib import admin

# Register your models here.
from .models import *


@admin.register(WhoopUser)
class WhoopUserAdmin(admin.ModelAdmin):
    list_display = ('user', 'whoop_user_id', 'access_token_updated', 'whoop_createdAt')
    search_fields = ('user__username', 'whoop_user_id')
    readonly_fields = ('access_token_updated', 'whoop_createdAt')


@admin.register(Daily)
class DailyAdmin(admin.ModelAdmin):
    list_display = ('id', 'user', 'day', 'lastUpdatedAt')
    list_filter = ('user', 'day')
    search_fields = ('id', 'user__user__username')
    date_hierarchy = 'day'


@admin.register(Recovery)
class RecoveryAdmin(admin.ModelAdmin):
    list_display = ('id', 'day', 'score', 'restingHeartRate', 'heartRateVariabilityRmssd', 'timestamp')
    list_filter = ('day__user',)
    search_fields = ('id', 'day__id')


@admin.register(Sleep)
class SleepAdmin(admin.ModelAdmin):
    list_display = ('id', 'day', 'score', 'qualityDuration', 'needBreakdown_total')
    list_filter = ('day__user',)
    search_fields = ('id', 'day__id')


@admin.register(SleepDetail)
class SleepDetailAdmin(admin.ModelAdmin):
    list_display = ('id', 'sleep', 'score', 'during_lower', 'during_upper', 'qualityDuration', 'isNap')
    list_filter = ('isNap', 'sleep__day__user')
    search_fields = ('id', 'sleep__id')
    date_hierarchy = 'during_lower'


@admin.register(Strain)
class StrainAdmin(admin.ModelAdmin):
    list_display = ('id', 'day', 'score', 'averageHeartRate', 'maxHeartRate', 'kilojoules')
    list_filter = ('day__user',)
    search_fields = ('id', 'day__id')


@admin.register(Workout)
class WorkoutAdmin(admin.ModelAdmin):
    list_display = ('id', 'strain', 'score', 'sportId', 'during_lower', 'during_upper', 'averageHeartRate')
    list_filter = ('sportId', 'strain__day__user')
    search_fields = ('id', 'strain__id')
    date_hierarchy = 'during_lower'


@admin.register(HR)
class HRAdmin(admin.ModelAdmin):
    list_display = ('id', 'user', 'timestamp', 'value')
    list_filter = ('user',)
    search_fields = ('user__user__username',)
    date_hierarchy = 'timestamp'


@admin.register(JournalEntry)
class JournalEntryAdmin(admin.ModelAdmin):
    list_display = ('id', 'day', 'title', 'category', 'answered_yes', 'entry_created_at')
    list_filter = ('category', 'answered_yes', 'day__user')
    search_fields = ('title', 'question_text', 'day__id')



"""
Tests for django_whoop models.

WHOOP API calls (requests.post / requests.get) are mocked throughout so no
network access is required.
"""
import datetime
from unittest.mock import MagicMock, patch

import pytz
import pytest
from django.contrib.auth.models import User

from django_whoop.models import (
    Daily,
    HR,
    JournalEntry,
    Recovery,
    Sleep,
    SleepDetail,
    Strain,
    WhoopUser,
    Workout,
    get_date,
    get_utc_time,
)


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def make_user(username="testuser"):
    return User.objects.create_user(username=username, password="secret")


def make_whoop_user(django_user=None):
    if django_user is None:
        django_user = make_user()
    return WhoopUser.objects.create(
        user=django_user,
        access_token="tok_access",
        refresh_token="tok_refresh",
        whoop_user_id=42,
    )


def make_daily(whoop_user, day_str="2021-03-17", daily_id=1001):
    return Daily.objects.create(
        id=daily_id,
        user=whoop_user,
        day=datetime.date.fromisoformat(day_str),
    )


# ---------------------------------------------------------------------------
# get_date utility
# ---------------------------------------------------------------------------

@pytest.mark.django_db
class TestGetDate:
    def test_none_returns_none(self):
        assert get_date(None) is None

    def test_iso_date(self):
        d = get_date("2021-03-17")
        assert d == datetime.datetime(2021, 3, 17)

    def test_iso_datetime_z(self):
        d = get_date("2021-03-17T12:00:00Z")
        assert d == datetime.datetime(2021, 3, 17, 12, 0, 0)

    def test_iso_datetime_with_millis(self):
        d = get_date("2021-03-17T12:00:00.000Z")
        assert d == datetime.datetime(2021, 3, 17, 12, 0, 0)

    def test_tzinfo_applied(self):
        d = get_date("2021-03-17", tzinfo=pytz.UTC)
        assert d.tzinfo == pytz.UTC

    def test_year_only(self):
        d = get_date("2021")
        assert d == datetime.datetime(2021, 1, 1)


# ---------------------------------------------------------------------------
# get_utc_time utility
# ---------------------------------------------------------------------------

@pytest.mark.django_db
class TestGetUtcTime:
    def test_none_returns_none(self):
        assert get_utc_time(None) is None

    def test_epoch_zero(self):
        d = get_utc_time(0)
        assert d == datetime.datetime(1970, 1, 1, tzinfo=pytz.UTC)

    def test_milliseconds_string(self):
        # 1_000_000_000 seconds since epoch — will fail without /1000, but the
        # function should handle a value already in seconds fine.
        ts = 1_616_000_000  # seconds
        d = get_utc_time(ts)
        assert d == datetime.datetime.utcfromtimestamp(ts).replace(tzinfo=pytz.UTC)

    def test_float_string(self):
        d = get_utc_time("1616000000.0")
        assert d is not None
        assert d.tzinfo == pytz.UTC


# ---------------------------------------------------------------------------
# WhoopUser model
# ---------------------------------------------------------------------------

@pytest.mark.django_db
class TestWhoopUser:
    def test_str(self):
        wu = make_whoop_user(make_user("alice"))
        assert str(wu) == "alice"

    def test_fields(self):
        user = make_user("bob")
        wu = make_whoop_user(user)
        assert wu.access_token == "tok_access"
        assert wu.refresh_token == "tok_refresh"
        assert wu.whoop_user_id == 42

    def test_one_to_one_with_user(self):
        user = make_user("carol")
        wu = make_whoop_user(user)
        assert user.whoopuser == wu

    @patch("django_whoop.models.requests.post")
    def test_create_with_password(self, mock_post):
        mock_resp = MagicMock()
        mock_resp.json.return_value = {
            "access_token": "new_access",
            "expires_in": 3600,
            "refresh_token": "new_refresh",
            "user": {
                "id": 99,
                "createdAt": "2020-10-18T00:57:18.609Z",
            },
        }
        mock_post.return_value = mock_resp

        django_user = make_user("dave")
        wu = WhoopUser.createWithPassword(django_user, "dave@example.com", "hunter2")
        assert wu.access_token == "new_access"
        assert wu.whoop_user_id == 99
        assert wu.refresh_token == "new_refresh"
        mock_post.assert_called_once()

    @patch("django_whoop.models.requests.post")
    def test_refresh_token(self, mock_post):
        mock_resp = MagicMock()
        mock_resp.json.return_value = {
            "access_token": "refreshed_access",
            "expires_in": 3600,
        }
        mock_post.return_value = mock_resp

        wu = make_whoop_user(make_user("eve"))
        wu.refreshToken()

        assert wu.access_token == "refreshed_access"
        mock_post.assert_called_once()

    @patch("django_whoop.models.requests.post")
    def test_refresh_with_password(self, mock_post):
        mock_resp = MagicMock()
        mock_resp.json.return_value = {
            "access_token": "pw_refreshed",
            "expires_in": 7200,
            "refresh_token": "new_rt",
        }
        mock_post.return_value = mock_resp

        wu = make_whoop_user(make_user("frank"))
        wu.refreshWithPassword("frank@example.com", "password123")

        assert wu.access_token == "pw_refreshed"
        assert wu.refresh_token == "new_rt"


# ---------------------------------------------------------------------------
# Daily model
# ---------------------------------------------------------------------------

@pytest.mark.django_db
class TestDailyModel:
    def test_str(self):
        wu = make_whoop_user()
        daily = make_daily(wu, "2021-03-17")
        assert str(daily) == "2021-03-17"

    def test_fields(self):
        wu = make_whoop_user()
        daily = make_daily(wu, "2021-03-17", daily_id=2001)
        assert daily.id == 2001
        assert daily.user == wu
        assert daily.day == datetime.date(2021, 3, 17)

    def test_fill_from_cycle_response_minimal(self):
        django_user = make_user("grace")
        wu = make_whoop_user(django_user)

        # Attach the reverse relation that fill_from_cycle_response expects
        # (user.whoopuser) — the view passes the django User object.
        django_user.whoopuser = wu

        cycle = {
            "id": 3001,
            "days": ["2021-03-17"],
            "during": {"bounds": "[)", "lower": "2021-03-17T00:00:00Z", "upper": "2021-03-18T00:00:00Z"},
            "lastUpdatedAt": "2021-03-17T20:00:00Z",
        }
        daily = Daily.fill_from_cycle_response(django_user, cycle)
        assert daily.id == 3001
        # get_date returns a datetime; the DateField stores/returns it as a date
        assert str(daily.day) == "2021-03-17"
        assert daily.during_bounds == "[)"

    def test_fill_from_cycle_response_with_recovery(self):
        django_user = make_user("hannah")
        wu = make_whoop_user(django_user)
        django_user.whoopuser = wu

        cycle = {
            "id": 3002,
            "days": ["2021-03-18"],
            "recovery": {
                "id": 60001,
                "heartRateVariabilityRmssd": 0.059,
                "restingHeartRate": 51,
                "score": 54,
                "timestamp": "2021-03-18T11:33:00Z",
            },
        }
        daily = Daily.fill_from_cycle_response(django_user, cycle)
        recovery = Recovery.objects.get(day=daily)
        assert recovery.score == 54
        assert recovery.restingHeartRate == 51

    def test_fill_from_cycle_response_with_strain_and_workout(self):
        django_user = make_user("ivan")
        wu = make_whoop_user(django_user)
        django_user.whoopuser = wu

        cycle = {
            "id": 3003,
            "days": ["2021-03-19"],
            "strain": {
                "averageHeartRate": 75,
                "kilojoules": 14218.7,
                "maxHeartRate": 172,
                "rawScore": 0.0143,
                "score": 16.39,
                "workouts": [
                    {
                        "id": 9001,
                        "averageHeartRate": 131,
                        "cumulativeWorkoutStrain": 14.9,
                        "during": {"bounds": "[)", "lower": "2021-03-19T16:53:55Z", "upper": "2021-03-19T18:08:15Z"},
                        "kilojoules": 2863.5,
                        "maxHeartRate": 157,
                        "rawScore": 0.0057,
                        "score": 12.24,
                        "sportId": 0,
                        "timezoneOffset": "-0400",
                        "zones": [100, 200, 300, 400, 500, 600],
                    }
                ],
            },
        }
        daily = Daily.fill_from_cycle_response(django_user, cycle)
        strain = Strain.objects.get(day=daily)
        assert strain.score == pytest.approx(16.39)
        workout = Workout.objects.get(strain=strain)
        assert workout.sportId == 0
        assert workout.zone_0 == 100
        assert workout.zone_5 == 600

    def test_fill_from_cycle_response_with_sleep(self):
        django_user = make_user("judy")
        wu = make_whoop_user(django_user)
        django_user.whoopuser = wu

        cycle = {
            "id": 3004,
            "days": ["2021-03-20"],
            "sleep": {
                "id": 7001,
                "needBreakdown": {
                    "baseline": 27430627,
                    "debt": 3703134,
                    "naps": 0,
                    "strain": 3278169,
                    "total": 34411932,
                },
                "qualityDuration": 25139744,
                "score": 73,
                "sleeps": [
                    {
                        "id": 8001,
                        "cyclesCount": 5,
                        "disturbanceCount": 12,
                        "during": {"bounds": "[)", "lower": "2021-03-20T03:01:00Z", "upper": "2021-03-20T11:33:00Z"},
                        "inBedDuration": 30718493,
                        "latencyDuration": 0,
                        "lightSleepDuration": 12011596,
                        "noDataDuration": 0,
                        "qualityDuration": 25139744,
                        "remSleepDuration": 5948524,
                        "respiratoryRate": 14.3262,
                        "score": 73,
                        "sleepConsistency": 89,
                        "sleepEfficiency": 0.818391,
                        "slowWaveSleepDuration": 7179624,
                        "timezoneOffset": "-0400",
                        "wakeDuration": 5615410,
                    }
                ],
            },
        }
        daily = Daily.fill_from_cycle_response(django_user, cycle)
        sleep = Sleep.objects.get(day=daily)
        assert sleep.score == 73
        assert sleep.needBreakdown_total == 34411932
        detail = SleepDetail.objects.get(sleep=sleep)
        assert detail.cyclesCount == 5
        assert detail.isNap is False


# ---------------------------------------------------------------------------
# Recovery model
# ---------------------------------------------------------------------------

@pytest.mark.django_db
class TestRecoveryModel:
    def test_create_and_retrieve(self):
        wu = make_whoop_user()
        daily = make_daily(wu)
        r = Recovery.objects.create(
            day=daily,
            id=60001,
            heartRateVariabilityRmssd=0.059,
            restingHeartRate=51,
            score=54,
        )
        fetched = Recovery.objects.get(pk=60001)
        assert fetched.score == 54
        assert fetched.day == daily


# ---------------------------------------------------------------------------
# Sleep / SleepDetail models
# ---------------------------------------------------------------------------

@pytest.mark.django_db
class TestSleepModels:
    def test_sleep_create(self):
        wu = make_whoop_user()
        daily = make_daily(wu)
        s = Sleep.objects.create(
            day=daily, id=7001, qualityDuration=25139744, score=73
        )
        assert Sleep.objects.filter(day=daily).count() == 1
        assert s.score == 73

    def test_sleep_detail_create(self):
        wu = make_whoop_user()
        daily = make_daily(wu)
        s = Sleep.objects.create(day=daily, id=7002, score=80)
        sd = SleepDetail.objects.create(
            sleep=s,
            id=8001,
            cyclesCount=5,
            isNap=False,
            respiratoryRate=14.3,
            sleepEfficiency=0.82,
        )
        assert sd.isNap is False
        assert sd.respiratoryRate == pytest.approx(14.3)


# ---------------------------------------------------------------------------
# Strain / Workout models
# ---------------------------------------------------------------------------

@pytest.mark.django_db
class TestStrainWorkout:
    def test_strain_create(self):
        wu = make_whoop_user()
        daily = make_daily(wu)
        s = Strain.objects.create(
            day=daily, id=5001, score=16.39, kilojoules=14218.7
        )
        assert s.score == pytest.approx(16.39)

    def test_workout_create(self):
        wu = make_whoop_user()
        daily = make_daily(wu)
        strain = Strain.objects.create(day=daily, id=5002, score=12.0)
        w = Workout.objects.create(
            strain=strain,
            id=9001,
            sportId=0,
            zone_0=100,
            zone_1=200,
            zone_2=300,
            zone_3=400,
            zone_4=500,
            zone_5=600,
        )
        assert w.zone_5 == 600
        assert w.sportId == 0


# ---------------------------------------------------------------------------
# HR model
# ---------------------------------------------------------------------------

@pytest.mark.django_db
class TestHRModel:
    def test_fill_from_response(self):
        wu = make_whoop_user()
        # Epoch ms for 2021-03-17T00:00:00Z → 1616025600000
        points = [
            {"time": 1616025600000, "data": 65},
            {"time": 1616025606000, "data": 66},
        ]

        # fill_from_response expects user.whoopuser
        django_user = wu.user
        django_user.whoopuser = wu

        HR.fill_from_response(user=django_user, d=points)
        assert HR.objects.filter(user=wu).count() == 2
        values = set(HR.objects.filter(user=wu).values_list("value", flat=True))
        assert values == {65, 66}


# ---------------------------------------------------------------------------
# JournalEntry model
# ---------------------------------------------------------------------------

@pytest.mark.django_db
class TestJournalEntry:
    def _make_daily(self):
        wu = make_whoop_user(make_user("journal_user"))
        return make_daily(wu, daily_id=4001)

    def test_str(self):
        daily = self._make_daily()
        entry = JournalEntry.objects.create(
            day=daily,
            id=500001,
            tracker_id=50,
            title="Intermittent Fasting",
            answered_yes=False,
        )
        assert "2021-03-17" in str(entry)
        assert "Intermittent Fasting" in str(entry)

    def test_create_from_response(self):
        daily = self._make_daily()
        raw = {
            "behavior_tracker": {
                "id": 50,
                "created_at": 1583441766.111,
                "updated_at": 1585669607.122,
                "title": "Intermittent Fasting",
                "status": "active",
                "category": "Nutrition",
                "description": None,
                "question_text": "Follow an intermittent fasting diet?",
                "time_label": "When was your last meal?",
                "time_context_label": "stopped eating at",
                "magnitude": None,
                "default_tracker": False,
            },
            "tracker_input": {
                "id": 500002,
                "journal_entry_id": 52336709,
                "created_at": 1611923762.42,
                "updated_at": 1611923762.42,
                "answered_yes": False,
                "behavior_tracker_id": 50,
                "time_input_value": None,
                "time_input_label": None,
                "magnitude_input_label": None,
                "magnitude_input_value": None,
            },
        }
        entry = JournalEntry.create_from_response(cycle_id=4001, d=raw)
        assert entry.title == "Intermittent Fasting"
        assert entry.answered_yes is False
        assert entry.tracker_id == 50
